#include <fcntl.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "serial.h"
#include "main.h"

#ifndef ARRAY_SIZE
#  define ARRAY_SIZE(x)	((sizeof (x))/(sizeof ((x)[0])))
#endif

#define FWDUMP_VERSION	"fwdump-0.0.2"


#define DEFAULT_TARGET	"*invalid_target*"
#define DEFAULT_DEVICE	"*invalid_device*"
#define DEFAULT_SPEED	"9600"
#define DEFAULT_BITS	"8"
#define DEFAULT_PARITY	"N"
#define DEFAULT_STOP	"1"
#define DEFAULT_OUTFILE	"rom.dump"
#define DEFAULT_START	"0x00000000"
#define DEFAULT_LEN	"0"
#define DEFAULT_SCHEME	"3"
#define DEFAULT_VERBOSE	0
#define DEFAULT_APPEND	0


static struct backend_descriptor {
	unsigned char *name;
	int (*init) (int fd);
	int (*get_byte) (int fd, unsigned char *output, unsigned long long int address);
	int (*fini) (int fd);
	int wordsize;
} backend[] = {
	{
		.name		= "vax",
		.wordsize	= 32,
		.init		= &vax_init,
		.get_byte	= &vax_get_byte,
	},
	{
		.name		= "alpha",
		.wordsize	= 64,
		.init		= &alpha_init,
		.get_byte	= &alpha_get_byte,
	},
};


struct backend_descriptor *
get_backend (unsigned char *name)
{
	int i;

	for (i = 0; i < ARRAY_SIZE (backend); i++)
		if (strcmp (name, backend[i].name) == 0)
			return &backend[i];

	return NULL;
}

static void
help (unsigned char *program_name)
{
	int i;

	fprintf (stderr, "Firmware dumper: %s\n", FWDUMP_VERSION);
	fprintf (stderr, "%s -- read area of memory through firmware\n", program_name);
	fprintf (stderr, "\n");
	fprintf (stderr, "--target xxx\n");
	fprintf (stderr, "--device /dev/ttySn\n");
	fprintf (stderr, "--speed nnnn\n");
	fprintf (stderr, "--bits {5,6,7,8}\n");
	fprintf (stderr, "--parity {E,O,N}\n");
	fprintf (stderr, "--stop {1,2}\n");
	fprintf (stderr, "--outfile xxxx\n");
	fprintf (stderr, "--start 0xNNNNNNNN\n");
	fprintf (stderr, "--len nnnn\n");
	fprintf (stderr, "--scheme n\n");
	fprintf (stderr, "--verbose\n");
	fprintf (stderr, "--append\n");
	fprintf (stderr, "\n");
	fprintf (stderr, "scheme:");
	fprintf (stderr, "\t1 - XXXXXXnn (low byte of byte longword)\n");
	fprintf (stderr, "\t2 - XXXXnnnn (low word of 4 byte longword)\n");
	fprintf (stderr, "\t3 - nnnnnnnn (full 4 byte longwords)\n");
	fprintf (stderr, "\t4 - XXXXXXXXXXXXXXnn (low byte of 8 byte longword)\n");
	fprintf (stderr, "\t5 - XXXXXXXXXXXXnnnn (low word of 8 byte longword)\n");
	fprintf (stderr, "\t6 - XXXXXXXXnnnnnnnn (low 4 byte longword of 8 byte longword)\n");
	fprintf (stderr, "\n");

	fprintf (stderr, "backends:\t");
	for (i = 0; i < ARRAY_SIZE (backend); i++) {
		if (i == 0)
			fprintf (stderr, "%s\n", backend[i].name);
		else
			fprintf (stderr, "\t\t%s\n", backend[i].name);
	}

	exit (EXIT_FAILURE);
}

static int
set_option (unsigned char **store, char *option)
{
	if (*store) {
		free (*store);
		*store = NULL;
	}

	printf ("Processing option \"%s\"\n", option);
	*store = strdup (option);

	return (*store == NULL)? -1: 0;
}

static int
read_memory_byte (struct backend_descriptor *target, int serial_fd,
		int outfile_fd, unsigned long long int address, unsigned char *data)
{
	int ret1;
	unsigned char one_byte;
	ssize_t ret2;

	ret1 = target->get_byte (serial_fd, &one_byte, address);
	if (ret1 != 0) {
		fprintf (stderr, "Failed to read byte at address 0x%0*llx\n",
				target->wordsize / 4, address);
		return -1;
	}

	ret2 = write (outfile_fd, &one_byte, sizeof (one_byte));
	if (ret2 != sizeof (one_byte)) {
		fprintf (stderr, "Failed to write byte at address %lld\n",
				address);
		return -2;
	}

	if (data)
		*data = one_byte;

	return 0;
}

static int
read_memory (struct backend_descriptor *target, int serial_fd,
		unsigned char *outfile, unsigned long long start,
		size_t len, int scheme, int verbose, int append)
{
	int outfile_fd;
	size_t i;
	int ret;
	int read_this_address;
	unsigned char last_byte;

	if (scheme < 1 || scheme > 6) {
		fprintf (stderr, "Don't know how to handle scheme %d\n", scheme);
		return -1;
	}

	outfile_fd = open (outfile, O_WRONLY | O_CREAT | O_NOCTTY |
			(append? O_APPEND: O_TRUNC),
			S_IRUSR | S_IWUSR | S_IRGRP);
	if (outfile_fd < 0) {
		perror ("Failed to open output file");
		return -2;
	}

	for (i = 0; i < len; i++) {
		read_this_address = 0;

		switch (scheme) {
			case 1:
				if (i % 4 == 0)
					read_this_address = 1; break;
			case 2:
				if (i % 4 == 0 || i % 4 == 1)
					read_this_address = 1; break;
			case 3:
				read_this_address = 1; break;
			case 4:
				if (i % 8 == 0)
					read_this_address = 1; break;
			case 5:
				if (i % 8 == 0 || i % 8 == 1)
					read_this_address = 1; break;
			case 6:
				if (i % 8 == 0 || i % 8 == 1 || i % 8 == 2 || i % 8 == 3)
					read_this_address = 1; break;
		}

		if (read_this_address) {
			ret = read_memory_byte (target, serial_fd, outfile_fd, start + i, &last_byte);
			if (ret != 0) {
				fprintf (stderr, "Failed to read byte at offset 0x%llx\n",
						(unsigned long long int) i);
				return -3;
			}
			if (verbose)
				printf ("[0x%0*llx] = [0x%0*llx+0x%0*x] = 0x%02x\n", target->wordsize / 4, start + i, target->wordsize / 4, start, target->wordsize / 4, i, last_byte);
		}
	}

	close (outfile_fd);

	return 0;
}

int
main (int argc, char *argv[])
{
	/* Variables we work with */
	struct backend_descriptor *back;
	int fd;
	int ret;

	int c;
	int show_help = 0;
	

	/* Variables the user supplied */
	unsigned char *target_string = strdup (DEFAULT_TARGET);
	unsigned char *start_string = strdup (DEFAULT_START);
	unsigned char *len_string = strdup (DEFAULT_LEN);
	unsigned char *outfile_string = strdup (DEFAULT_OUTFILE);
	unsigned char *scheme_string = strdup (DEFAULT_SCHEME);
	unsigned char *device_string = strdup (DEFAULT_DEVICE);
	unsigned char *speed_string = strdup (DEFAULT_SPEED);
	unsigned char *bits_string = strdup (DEFAULT_BITS);
	unsigned char *parity_string = strdup (DEFAULT_PARITY);
	unsigned char *stop_string = strdup (DEFAULT_STOP);
	int verbose = DEFAULT_VERBOSE;
	int append = DEFAULT_APPEND;

	while (1) {
		int option_index = 0;
		static struct option long_options[] = {
			{ .name	= "target",	.has_arg = required_argument,	.flag = NULL, .val = 'a', },
			{ .name	= "start",	.has_arg = required_argument,	.flag = NULL, .val = 'b', },
			{ .name	= "len",	.has_arg = required_argument,	.flag = NULL, .val = 'c', },
			{ .name = "device",	.has_arg = required_argument,	.flag = NULL, .val = 'd', },
			{ .name = "speed",	.has_arg = required_argument,	.flag = NULL, .val = 'e', },
			{ .name = "bits",	.has_arg = required_argument,	.flag = NULL, .val = 'f', },
			{ .name = "parity",	.has_arg = required_argument,	.flag = NULL, .val = 'g', },
			{ .name = "stop",	.has_arg = required_argument,	.flag = NULL, .val = 'h', },
			{ .name = "help",	.has_arg = no_argument,		.flag = NULL, .val = 'i', },
			{ .name = "outfile",	.has_arg = required_argument,	.flag = NULL, .val = 'j', },
			{ .name = "scheme",	.has_arg = required_argument,	.flag = NULL, .val = 'k', },
			{ .name = "verbose",	.has_arg = no_argument,		.flag = NULL, .val = 'v', },/**/
			{ .name = "append",	.has_arg = no_argument,		.flag = NULL, .val = 'l', },
		};

		c = getopt_long (argc, argv, "a:b:c:d:e:f:g:h:ivl", long_options, &option_index);
		if (c == -1)
			break;

		switch (c) {
			case 'a':
				if (set_option (&target_string, optarg)) {
					fprintf (stderr, "Failed to set target\n");
					exit (EXIT_FAILURE);
				}
				break;

			case 'b':
				if (set_option (&start_string, optarg)) {
					fprintf (stderr, "Failed to set start address\n");
					exit (EXIT_FAILURE);
				}
				break;

			case 'c':
				if (set_option (&len_string, optarg)) {
					fprintf (stderr, "Failed to set len to dump\n");
					exit (EXIT_FAILURE);
				}
				break;

			case 'd':
				if (set_option (&device_string, optarg)) {
					fprintf (stderr, "Failed to set device name\n");
					exit (EXIT_FAILURE);
				}
				break;

			case 'e':
				if (set_option (&speed_string, optarg)) {
					fprintf (stderr, "Failed to set speed\n");
					exit (EXIT_FAILURE);
				}
				break;
			case 'f':
				if (set_option (&bits_string, optarg)) {
					fprintf (stderr, "Failed to set number of bits\n");
					exit (EXIT_FAILURE);
				}
				break;

			case 'g':
				if (set_option (&parity_string, optarg)) {
					fprintf (stderr, "Failed to set parity\n");
					exit (EXIT_FAILURE);
				}
				break;

			case 'h':
				if (set_option (&stop_string, optarg)) {
					fprintf (stderr, "Failed to set number of stop-bits\n");
					exit (EXIT_FAILURE);
				}
				break;

			case 'j':
				if (set_option (&outfile_string, optarg)) {
					fprintf (stderr, "Failed to set name of output file\n");
					exit (EXIT_FAILURE);
				}
				break;

			case 'k':
				if (set_option (&scheme_string, optarg)) {
					fprintf (stderr, "Failed to set byte gathering scheme\n");
					exit (EXIT_FAILURE);
				}
				break;

			case 'v':
				verbose = 1;
				break;

			case 'l':
				append = 1;
				break;

			case 'i':
			default:
				show_help = 1;
				break;
		}
	}

	if (show_help) {
		help (argv[0]);
		return EXIT_FAILURE;
	}

	back = get_backend (target_string);
	if (!back) {
		fprintf (stderr, "Didn't find target \"%s\"\n", target_string);
		return EXIT_FAILURE;
	}

	fd = serial_open_luxury (
			device_string,
			atoi (speed_string),
			atoi (bits_string),
			(parity_string && parity_string[0])?
				*parity_string:
				DEFAULT_PARITY[0],
			atoi (stop_string));
	if (fd < 0) {
		fprintf (stderr, "Failed to open serial interface \"%s\"\n",
				device_string);
		exit (EXIT_FAILURE);
	}

	if (back->init)
		back->init (fd);

	ret = read_memory (back, fd, outfile_string,
			strtoull (start_string, NULL, 0),
			strtoul (len_string, NULL, 0),
			atoi (scheme_string), verbose, append);
	if (back->fini)
		back->fini (fd);

	if (ret != 0) {
		fprintf (stderr, "Failed to read memory\n");
		close (fd);
		return EXIT_FAILURE;
	}

	close (fd);

	return EXIT_SUCCESS;
}

