#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include "serial.h"

int
serial_open (unsigned char *devname)
{
	int fd;

	fd = open (devname, O_RDWR | O_NOCTTY | O_NONBLOCK);
	if (fd < 0)
		return -1;

	return fd;
}

int
serial_close (int fd)
{
	close (fd);

	return 0;
}

int
serial_set_bits (int fd, unsigned char bits)
{
	struct termios termios;
	int termios_bits;
	int ret;

	ret = tcgetattr (fd, &termios);
	if (ret != 0)
		return -1;

	termios.c_cflag = termios.c_cflag & ~CSIZE;
	switch (bits) {
		case 5:
			termios_bits = CS5;
			break;

		case 6:
			termios_bits = CS6;
			break;

		case 7:
			termios_bits = CS7;
			break;

		case 8:
			termios_bits = CS8;
			break;

		default:
			termios_bits = CS8;
			break;
	}
	termios.c_cflag |= termios_bits;

	ret = tcsetattr (fd, TCSANOW, &termios);
	if (ret != 0)
		return -2;

	return 0;
}

int
serial_set_parity (int fd, unsigned char parity)
{
	struct termios termios;
	int ret;

	ret = tcgetattr (fd, &termios);
	if (ret != 0)
		return -1;

	switch (parity) {
		case 'E':
		case 'e':
			termios.c_cflag &= ~PARODD;
			termios.c_cflag |= PARENB;
			break;

		case 'O':
		case 'o':
			termios.c_cflag |= ~PARODD;
			termios.c_cflag |= PARENB;
			break;

		case 'N':
		case 'n':
			termios.c_cflag &= ~PARODD;
			termios.c_cflag &= ~PARENB;
			break;

		default:
			termios.c_cflag &= ~PARODD;
			termios.c_cflag &= ~PARENB;
			break;
	}

	ret = tcsetattr (fd, TCSANOW, &termios);
	if (ret != 0)
		return -2;

	return 0;
}

int
serial_set_stop (int fd, unsigned char stop)
{
	struct termios termios;
	int ret;

	ret = tcgetattr (fd, &termios);
	if (ret != 0)
		return -1;

	switch (stop) {
		case 1:
			termios.c_cflag &= ~CSTOPB;
			break;

		case 2: 
			termios.c_cflag |= CSTOPB;
			break;

		default:
			termios.c_cflag &= ~CSTOPB;
			break;
	}

	ret = tcsetattr (fd, TCSANOW, &termios);
	if (ret != 0)
		return -2;

	return 0;
}

int
serial_set_one_byte_each (int fd)
{
	struct termios termios;
	int ret;

	ret = tcgetattr (fd, &termios);
	if (ret != 0)
		return -1;

	termios.c_iflag |= IGNPAR;	/* Ignore bytes with parity errors */
	termios.c_cflag |= CLOCAL;	/* No modem control lines */
	termios.c_cflag |= CREAD;	/* Enable serial receiver */
	termios.c_cflag &= ~CRTSCTS;	/* Disable handshake */
	termios.c_oflag = 0;		/* Raw output */
	termios.c_lflag = 0;		/* Raw local handling */
	termios.c_cc[VTIME] = 0;	/* Inter-character timer unused */
	termios.c_cc[VMIN] = 1;		/* One byte per read */
	
	ret = tcsetattr (fd, TCSANOW, &termios);
	if (ret != 0)
		return -2;

	return 0;
}

int
serial_set_speed (int fd, unsigned int speed)
{
	struct termios termios;
	unsigned int termios_speed;
	int ret;

	ret = tcgetattr (fd, &termios);
	if (ret != 0)
		return -1;

	termios.c_cflag = termios.c_cflag & ~CBAUD;
	termios.c_cflag = termios.c_cflag & ~CBAUDEX;
	switch (speed) {
#ifdef B0
		case 0:		termios_speed = B0;		break;
#endif
#ifdef B50
		case 50:	termios_speed = B50;		break;
#endif
#ifdef B75
		case 75:	termios_speed = B75;		break;
#endif
#ifdef B110
		case 110:	termios_speed = B110;		break;
#endif
#ifdef B134
		case 134:	termios_speed = B134;		break;
#endif
#ifdef B150
		case 150:	termios_speed = B150;		break;
#endif
#ifdef B200
		case 200:	termios_speed = B200;		break;
#endif
#ifdef B300
		case 300:	termios_speed = B300;		break;
#endif
#ifdef B600
		case 600:	termios_speed = B600;		break;
#endif
#ifdef B1200
		case 1200:	termios_speed = B1200;		break;
#endif
#ifdef B1800
		case 1800:	termios_speed = B1800;		break;
#endif
#ifdef B2400
		case 2400:	termios_speed = B2400;		break;
#endif
#ifdef B4800
		case 4800:	termios_speed = B4800;		break;
#endif
#ifdef B9600
		case 9600:	termios_speed = B9600;		break;
#endif
#ifdef B19200
		case 19200:	termios_speed = B19200;		break;
#endif
#ifdef B38400
		case 38400:	termios_speed = B38400;		break;
#endif
#ifdef B57600
		case 57600:	termios_speed = B57600;		break;
#endif
#ifdef B115200
		case 115200:	termios_speed = B115200;	break;
#endif
#ifdef B230400
		case 230400:	termios_speed = B230400;	break;
#endif
#ifdef B460800
		case 460800:	termios_speed = B460800;	break;
#endif
#ifdef B5000000
		case 500000:	termios_speed = B500000;	break;
#endif
#ifdef B576000
		case 576000:	termios_speed = B576000;	break;
#endif
#ifdef B921600
		case 921600:	termios_speed = B921600;	break;
#endif
#ifdef B1000000
		case 1000000:	termios_speed = B1000000;	break;
#endif
#ifdef B1152000
		case 1152000:	termios_speed = B1152000;	break;
#endif
#ifdef B1500000
		case 1500000:	termios_speed = B1500000;	break;
#endif
#ifdef B2000000
		case 2000000:	termios_speed = B2000000;	break;
#endif
#ifdef B2500000
		case 2500000:	termios_speed = B2500000;	break;
#endif
#ifdef B3000000
		case 3000000:	termios_speed = B3000000;	break;
#endif
#ifdef B3500000
		case 3500000:	termios_speed = B3500000;	break;
#endif
#ifdef B4000000
		case 4000000:	termios_speed = B4000000;	break;
#endif
		default:
				termios_speed = B9600;
				break;
	}
	termios.c_cflag |= termios_speed;

	ret = tcsetattr (fd, TCSANOW, &termios);
	if (ret != 0)
		return -2;

	return 0;
}

int
serial_write_byte (int fd, unsigned char byte)
{
	ssize_t ret;

	ret = write (fd, &byte, 1);
	if (ret != 1)
		return -1;

	return 0;
}

int
serial_write_byte_rb (int fd, unsigned char byte)
{
	ssize_t ret;
	int ret2;
	unsigned char readback_byte;

	ret = write (fd, &byte, 1);
	if (ret != 1)
		return -1;

	ret2 = serial_read_byte (fd, &readback_byte, 1, 0);
	if (ret2 != 0)
		return -2;
	else
		return 0;
}

size_t
serial_write (int fd, unsigned char *buf, size_t len)
{
	ssize_t ret;
	size_t written = 0;
	int errcount = 0;

	while (written < len) {
		ret = write (fd, buf + written, len - written);
		if (ret >= 0) {
			written += ret;
			errcount = 0;
		} else {
			switch (errno) {
				/* "Soft" errors allow retransmission */
				case EINTR:
				case ENOSPC:
				case EAGAIN:
					if (errcount++ > 200)
						return len - written;
					break;

				/* Anything else is considered fatal */
				default:
					return len - written;
					break;
			}
		}
	}

	return 0;
}

size_t
serial_write_rb (int fd, unsigned char *buf, size_t len)
{
	int i;
	size_t written = 0;
	int ret;

	for (i = 0; i < len; i++) {
		ret = serial_write_byte_rb (fd, buf[i]);
		if (ret == 0)
			written++;
		else
			break;
	}

	return len - written;
}

int
serial_read_byte (int fd, unsigned char *byte, time_t sec, suseconds_t usec)
{
	struct timeval timeout;
	fd_set read_set;
	unsigned char my_byte;
	int ret;
	ssize_t len;

	timeout.tv_sec = sec;
	timeout.tv_usec = usec;
	FD_ZERO (&read_set);
	FD_SET (fd, &read_set);

	ret = select (fd + 1, &read_set, NULL, NULL, &timeout);
	if (ret > 0) {
		len = read (fd, &my_byte, 1);
		if (len == 1) {
			*byte = my_byte;
			return 0;
		} else
			return -1;
	} else
		return -2;
}

int
serial_set_line_parameters (int fd, unsigned int speed, unsigned char bits,
		unsigned char parity, unsigned char stop)
{
	int ret;

	ret = serial_set_speed (fd, speed);
	if (ret != 0)
		return -1;

	ret = serial_set_bits (fd, bits);
	if (ret != 0)
		return -2;

	ret = serial_set_parity (fd, parity);
	if (ret != 0)
		return -3;

	ret = serial_set_stop (fd, stop);
	if (ret != 0)
		return -4;

	return 0;
}

int
serial_open_luxury (unsigned char *device, unsigned int speed,
		unsigned char bits, unsigned char parity, unsigned char stop)
{
	int fd;
	int ret;

	fd = serial_open (device);
	if (fd < 0)
		return -1;

	ret = serial_set_line_parameters (fd, speed, bits, parity, stop);
	if (ret != 0) {
		close (fd);
		return -2;
	}

	ret = serial_set_one_byte_each (fd);
	if (ret != 0) {
		close (fd);
		return -3;
	}

	return fd;
}

int
serial_get_modem (int fd, int what, int *result)
{
	int ret;
	int status;

	status = what;
	ret = ioctl (fd, TIOCMGET, &status);
	if (ret == 0) {
		*result = status;
		return 0;
	} else
		return -1;
}

int
serial_set_modem (int fd, int what, int state)
{
	int ret;
	int old_state;
	int new_state;

	ret = ioctl (fd, TIOCMGET, &old_state);
	if (ret != 0)
		return -1;

	if (state)
		new_state = old_state | what;
	else
		new_state = old_state & ~what;

	ret = ioctl (fd, TIOCMSET, &new_state);
	if (ret != 0)
		return -1;

	return 0;
}

int
serial_send_break (int fd, int duration)
{
	return tcsendbreak (fd, duration);
}

